#Offline diffraction simulation scripts from pisha2020approximate
#Copyright (C) 2020 Louis Pisha <lpisha@eng.ucsd.edu> / Sonic Arts Research & Development <sonicarts.ucsd.edu> / UC San Diego / The Regents of the University of California 
#See LICENSE.txt for licensing information

# 3D live-adjustable diffraction simulator. Includes GTD, UTD, BTM with
# low-quality parameters, BTM with high-quality parameters (WARNING, this
# freezes the interface for 10+ seconds while computing!), the VDaT ring
# transmission results, and the VDaT amplitude response. Also plots the path
# lengths computed from each approach.
# 
# Note that matplotlib's 3D plotting has some major bugs when attempting to plot
# geometry which is much larger than the figure, e.g. the half-plane and 
# infinite wedge. This is only a graphical glitch--the diffraction models are
# being computed correctly.
# 
# The radio (option) buttons in the third column control which diffraction model
# is being visualized in the 3D view. If this is set to VDaT, the check boxes
# in the fourth column control which VDaT rings are visible.
# The check boxes in the fifth column control which VDaT subpaths are in use.
# 
# Figure 13 in pisha2020approximate generated with this script.

#figsize = (5.0, 5.0)
figsize = (9.0, 9.0)
ylim = (-35.0, 15.0)

import math
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.tri as mtri
from matplotlib.widgets import Slider, CheckButtons, RadioButtons
from mpl_toolkits.mplot3d import Axes3D

from common import *
import examples3d
from xtd import gtd_core, utd_core, call_xtd
from btm import call_btm
from vdat import vdat

# Parameters

always = True
xtd_oob = True
mesh = 'Plane'
view_as = 'BTM'
show_direct = False
show_gtd = False
show_utd = True
show_btmlo = True
show_btmhi = False
show_vdatrxm = False
show_vdatmag = True

infsize = 20.0
zoom = 1.0
ps = [0.1, 0.2, -0.8]
pk = [-0.05, 0.12, 0.9]
meshscale = 0.27
angle = 90.0

c = 345.0
ringdefs = defaultringdefs(9, 64, 44100, c)
vdatparams = defaultvdatparams()
default_freqs = 2.0 * np.logspace(1, 4.2, num=50)

btmloparams = {'fs': 88200, 'c': c, 'irlen': 512, 'totalirlen': 16384, 'goodnumint': False}
btmhiparams = {'fs': 160000, 'c': c, 'irlen': 65536, 'totalirlen': 65536*16, 'goodnumint': True}

# Computation

def recompute():
    # Functions
    def points_array_to_xyz(points):
        return [[float(p[i]) for p in points] for i in range(3)]
    def plot_mesh(verts, tris):
        if len(tris) == 0: return
        tri = mtri.Triangulation([v[0] for v in verts], [v[1] for v in verts], tris)
        simax.plot_trisurf(tri, [v[2] for v in verts], lw=0.2, edgecolor='black', color='grey', alpha=0.5)
    def plot_edges(verts, edges):
        def edge_verts(e):
            return [[verts[e['i1']][d], verts[e['i2']][d]] for d in range(0,3)]
        def vcenter(v):
            return [(vv[0] + vv[1]) / 2.0 for vv in v]
        def vdiff(v):
            return [(vv[1] - vv[0]) for vv in v]
        for e in edges:
            v = edge_verts(e)
            simax.plot(*v, lw=3.0, c='black', marker='')
            #s = '{:.0f}°'.format(e['angle'] * 180.0 / math.pi)
            #simax.text(*vcenter(v), s, vdiff(v))
    def plot_xtd_path(apex, inedge):
        simax.plot(*points_array_to_xyz([ps, apex, pk]), lw=1.5, marker='', 
            c='darkgoldenrod' if inedge else 'lightgrey', linestyle='-' if inedge else '--')
        simax.plot([apex[0]], [apex[1]], [apex[2]], marker='^',
            c='black' if inedge else 'grey')
    def plot_btm_path(ev1, ev2):
        verts = [ev1, ev2, ps, pk]
        tris = [[0, 1, 2], [0, 1, 3]]
        tri = mtri.Triangulation([v[0] for v in verts], [v[1] for v in verts], tris)
        simax.plot_trisurf(tri, [v[2] for v in verts], color='red', alpha=0.3)
    def plot_vdat_sps(sps):
        for sp in sps:
            simax.plot(*points_array_to_xyz(sp['sp']), lw=1, marker='', 
                color='yellow' if sp['xm'] >= 0.5 else 'darkred', alpha=0.5)
    
    # Mesh
    if mesh == 'None':
        verts, tris, edges = examples3d.nomesh()
    elif mesh == 'Half-plane':
        verts, tris, edges = examples3d.halfplane(infsize)
    elif mesh == 'Inf wedge':
        verts, tris, edges = examples3d.infwedge(radians(angle), infsize)
    elif mesh == 'Plane':
        verts, tris, edges = examples3d.plane()
    elif mesh == 'Plane 1 ed.':
        verts, tris, edges = examples3d.plane_one_edge()
    elif mesh == 'Octagon':
        verts, tris, edges = examples3d.octagon()
    else:
        raise RuntimeError('Invalid mesh: ' + mesh)
    verts = [[w*meshscale for w in v] for v in verts]
    
    simax.clear()
    simax.set_xlim(-zoom, zoom)
    simax.set_ylim(-zoom, zoom)
    simax.set_zlim(-zoom, zoom)
    plot_mesh(verts, tris)
    if view_as in ['XTD', 'BTM']:
        plot_edges(verts, edges)
        
    for i, edge in enumerate(edges):
        #print('Edge ' + str(i) + ':')
        try:
            ri, rd, zi, zd, li, ld, phii, apex, inedge, zedgelo, zedgehi, thetai, thetad, v \
                = edge_params(verts, tris, edge, ps, pk)
            if view_as == 'XTD' and not inedge and not xtd_oob:
                #print('Out of bounds')
                continue
            if view_as == 'XTD':
                plot_xtd_path(apex, inedge)
            elif view_as == 'BTM':
                plot_btm_path(verts[edge['i1']], verts[edge['i2']])
        except ValueError as err:
            print(err)
    
    alllens = []
    freqs = default_freqs
    ground_truth = None
    if show_btmlo:
        btmlores, btmlofreqs, btmlolens_nz, btmlolens_me = call_btm(
            verts, tris, edges, ps, pk, always, btmloparams)
        btmlores = todb(audible_only(btmlofreqs, btmlores))
        freqs = audible_only(btmlofreqs)
        ground_truth = btmlores
        alllens.append({'l': btmlolens_nz, 'label': 'BTM (low)\nNonzero', 'color': 'cyan'})
        alllens.append({'l': btmlolens_me, 'label': 'BTM (low)\nMean E.', 'color': 'cyan'})
    if show_btmhi:
        btmhires, btmhifreqs, btmhilens_nz, btmhilens_me = call_btm(
            verts, tris, edges, ps, pk, always, btmhiparams)
        btmhires = todb(audible_only(btmhifreqs, btmhires))
        freqs = audible_only(btmhifreqs)
        ground_truth = btmhires
        alllens.append({'l': btmhilens_nz, 'label': 'BTM\nNonzero', 'color': 'green'})
        alllens.append({'l': btmhilens_me, 'label': 'BTM\nMean E.', 'color': 'green'})
    
    if show_direct:
        directlen = norm(np.array(pk) - np.array(ps))
        directres = [1.0 / directlen, 1.0 / directlen]
        directres = todb(directres)
        directfreqs = [1.0, 100000.0]
        directlens = [directlen]
        alllens.append({'l': directlens, 'label': 'Direct', 'color': 'darkgray'})
    if show_gtd:
        gtdres, gtdlens = call_xtd(gtd_core, verts, tris, edges, ps, pk, freqs, c, always, xtd_oob)
        gtdres = todb(gtdres)
        alllens.append({'l': gtdlens, 'label': 'GTD', 'color': 'red'})
    if show_utd:
        utdres, utdlens = call_xtd(utd_core, verts, tris, edges, ps, pk, freqs, c, always, xtd_oob)
        utdres = todb(utdres)
        alllens.append({'l': utdlens, 'label': 'UTD', 'color': 'blue'})
    if show_vdatrxm or show_vdatmag:
        vdatrxmres, vdatmagres, vdatlens, vdatsps = vdat(
            verts, tris, edges, ps, pk, ringdefs, freqs, always, vdatparams)
        vdatrxmres = todb(vdatrxmres)
        vdatmagres = todb(vdatmagres)
        alllens.append({'l': vdatlens, 'label': 'VDaT', 'color': 'darkorange'})
    
    if view_as == 'VDaT' and (show_vdatrxm or show_vdatmag):
        plot_vdat_sps(vdatsps)
    
    if ground_truth is not None:
        if show_utd:
            fwlsd = FW_LSD(freqs, ground_truth, utdres, True)
            print('UTD FW-LSD: {:.2f} dB'.format(fwlsd))
        if show_vdatmag:
            fwlsd = FW_LSD(freqs, ground_truth, vdatmagres, True)
            print('VDaT FW-LSD: {:.2f} dB'.format(fwlsd))
    
    simax.plot([ps[0]], [ps[1]], [ps[2]], c='lime', marker='o', markersize=12)
    simax.plot([pk[0]], [pk[1]], [pk[2]], c='blue', marker='o', markersize=12)
    
    magax.clear()
    magax.set_xscale('log')
    magax.set_xlim(20, 20000)
    magax.set_ylim(*ylim)
    ring_fs = [r['f'] for r in ringdefs]
    if show_direct:
        magax.plot(directfreqs, directres, marker='', color='darkgray', label='Direct')
    if show_gtd:
        magax.plot(freqs, gtdres, marker='', color='red', label='GTD')
    if show_utd:
        magax.plot(freqs, utdres, marker='', linestyle=':', color='blue', label='UTD')
    if show_btmlo:
        magax.plot(freqs if not show_btmhi else audible_only(btmlofreqs), btmlores, marker='', color='cyan', label='BTM (low)')
    if show_btmhi:
        magax.plot(freqs, btmhires, marker='', color='green', label='BTM')
    if show_vdatrxm:
        magax.plot(ring_fs, vdatrxmres, marker='.', color='purple', label='VDaT rxm')
    if show_vdatmag:
        magax.plot(freqs, vdatmagres, marker='', linestyle='--', color='darkorange', label='VDaT')
    magax.set_xlabel('Frequency (Hz)')
    magax.set_ylabel('Amplitude (dB)')
    magax.set_title('Total amplitude response')
    magax.legend(loc='lower left')
    
    lenax.clear()
    lenax.set_xticks(np.arange(len(alllens)))
    lenax.set_xticklabels([l['label'] for l in alllens])
    lenax.set_xlim((-1, len(alllens)))
    lenax.set_ylabel('Length (m)')
    set_width = 0.7
    for nl, l in enumerate(alllens):
        lens = l['l']
        w = set_width / len(lens)
        xs = np.arange(len(lens)) * w - (0.5*set_width) + nl
        lenax.bar(xs, lens, w, color=l['color'], align='edge', edgecolor='black', linewidth=1)
        for x, l in zip(xs, lens):
            lenax.text(x + 0.5*w, l, '{:.2f}'.format(l), ha='center', va='bottom')
            
    plt.gcf().canvas.draw_idle()

# Subplots

plt.figure(figsize=figsize, dpi=100)
simax = plt.axes((0.05, 0.55, 0.4, 0.4), projection='3d')
magax = plt.axes((0.55, 0.55, 0.4, 0.4))
lenax = plt.axes((0.55, 0.05, 0.4, 0.4))

# Controls

def sldupdate(val):
    global zoom, ps, pk, meshscale, angle
    zoom = pow(10.0, -sliders['Zoom'].val)
    ps[0] = zoom * sliders['X1'].val
    ps[1] = zoom * sliders['Y1'].val
    ps[2] = zoom * sliders['Z1'].val
    pk[0] = zoom * sliders['X2'].val
    pk[1] = zoom * sliders['Y2'].val
    pk[2] = zoom * sliders['Z2'].val
    meshscale = pow(10.0, -sliders['MshScl'].val)
    angle = sliders['Angle'].val
    recompute()

def make_slider(idx, label, col, type, init, y, sldupdate):
    if type == 'cart':
        mn, mx, step = -1.0, 1.0, 0.01
    elif type == 'scale':
        mn, mx, step = -2.0, 2.0, 0.01
    elif type == 'angle':
        mn, mx, step = 0.0, 180.0, 1.0
    else:
        raise ValueError('Invalid slider type')
    sldh = 0.02
    sld = Slider(plt.axes([0.05, y, 0.4, sldh], facecolor=col),
        label, mn, mx, valinit=init, valstep=step, facecolor=adjust_lightness(col, 0.3))
    sld.on_changed(sldupdate)
    return sld

sld_labels = ['Angle', 'MshScl', 'Z2', 'Y2', 'X2', 'Z1', 'Y1', 'X1', 'Zoom']
sld_cols = ['gold', 'purple', 'blue', 'lime', 'red', 'blue', 'lime', 'red', 'purple']
sld_types = ['angle', 'scale', 'cart', 'cart', 'cart', 'cart', 'cart', 'cart', 'scale']
sld_inits = [angle, -log10(meshscale), pk[2], pk[1], pk[0], ps[2], ps[1], ps[0], -log10(zoom)]
sld_ys = [0.05, 0.07, 0.10, 0.12, 0.14, 0.17, 0.19, 0.21, 0.24]
sliders = {}
i = 0
for label, col, type, init, y in zip(sld_labels, sld_cols, sld_types, sld_inits, sld_ys):
    sliders[label] = make_slider(i, label, col, type, init, y, sldupdate)
    i += 1

def chksChksUpdate(label):
    global always, xtd_oob
    if label == 'Always':
        always = not always
    elif label == 'XTD OOB':
        xtd_oob = not xtd_oob
    else:
        raise RuntimeError('Invalid chksChks label')
    recompute()

def optsMeshUpdate(label):
    global mesh
    mesh = label
    recompute()
    
def chksShowUpdate(label):
    global show_direct, show_gtd, show_utd, show_btmlo, show_btmhi, show_vdatrxm, show_vdatmag
    if label == 'Direct':
        show_direct = not show_direct
    elif label == 'GTD': 
        show_gtd = not show_gtd
    elif label == 'UTD': 
        show_utd = not show_utd
    elif label == 'BTM (low)':
        show_btmlo = not show_btmlo
    elif label == 'BTM':
        show_btmhi = not show_btmhi
    elif label == 'VDaT rxm':
        show_vdatrxm = not show_vdatrxm
    elif label == 'VDaT':
        show_vdatmag = not show_vdatmag
    else:
        raise RuntimeError('Invalid chksShow label')
    recompute()
chksShowLabels = ['Direct', 'GTD', 'UTD', 'BTM (low)', 'BTM', 'VDaT rxm', 'VDaT']
chksShowDefaults = [show_direct, show_gtd, show_utd, show_btmlo, show_btmhi, show_vdatrxm, show_vdatmag]

def optsViewAsUpdate(label):
    global view_as
    view_as = label
    recompute()
    
def chksBandUpdate(label):
    global ringdefs
    for f in ringdefs:
        if str(int(f['f'])) == label:
            f['showsps'] = not f['showsps']
            print('Setting ' + str(int(f['showsps'])) + ' to ' + str(f['showsps']))
    recompute()

def chksSubpathUpdate(label):
    global vdatparams
    vdatparams['subpath' + label] = not vdatparams['subpath' + label]
    recompute()

chksChks   = CheckButtons(plt.axes([0.02, 0.40, 0.10, 0.05]), ['Always', 'XTD OOB'], [always, xtd_oob])
optsMeshText = ['None', 'Half-plane', 'Inf wedge', 'Plane', 'Plane 1 ed.', 'Octagon']
optsMesh   = RadioButtons(plt.axes([0.02, 0.30, 0.10, 0.10]), optsMeshText, optsMeshText.index(mesh))
chksShow   = CheckButtons(plt.axes([0.12, 0.30, 0.10, 0.15]), chksShowLabels, chksShowDefaults)
optsViewAsText = ['None', 'XTD', 'BTM', 'VDaT']
optsViewAs = RadioButtons(plt.axes([0.22, 0.30, 0.07, 0.15]), optsViewAsText, optsViewAsText.index(view_as))
chksBand   = CheckButtons(plt.axes([0.29, 0.30, 0.07, 0.15]), 
    [str(int(r['f'])) for r in ringdefs], [r['showsps'] for r in ringdefs])
chksSubpath= CheckButtons(plt.axes([0.36, 0.30, 0.04, 0.15]), range(4), 
    [vdatparams['subpath' + str(i)] for i in range(4)])

fix_widget(chksChks, True)
fix_widget(optsMesh, False)
fix_widget(chksShow, True)
fix_widget(optsViewAs, False)
fix_widget(chksBand, True)
fix_widget(chksSubpath, True)

chksChks.on_clicked(chksChksUpdate)
optsMesh.on_clicked(optsMeshUpdate)
chksShow.on_clicked(chksShowUpdate)
optsViewAs.on_clicked(optsViewAsUpdate)
chksBand.on_clicked(chksBandUpdate)
chksSubpath.on_clicked(chksSubpathUpdate)

recompute()
plt.show()
