#Offline diffraction simulation scripts from pisha2020approximate
#Copyright (C) 2020 Louis Pisha <lpisha@eng.ucsd.edu> / Sonic Arts Research & Development <sonicarts.ucsd.edu> / UC San Diego / The Regents of the University of California 
#See LICENSE.txt for licensing information

# Example 3D geometry for testing diffraction.

import math
import numpy as np
from common import *

def mesh_edges(verts, tris, discard_cosangle = -0.9, discard_single = False):
    edges = []
    for ti, t in enumerate(tris):
        def add_edge(i1, i2, ti):
            for e in edges:
                if i1 == e['i1'] and i2 == e['i2'] or i2 == e['i1'] and i1 == e['i2']:
                    e['tis'].append(ti)
                    return
            edges.append({'i1': i1, 'i2': i2, 'tis': [ti]})
        add_edge(t[0], t[1], ti)
        add_edge(t[0], t[2], ti)
        add_edge(t[1], t[2], ti)
    for e in edges:
        tis = e['tis']
        if len(tis) == 1:
            if discard_single:
                e['tis'] = None
            else:
                e['angle'] = math.tau
        elif len(tis) == 2:
            i1, i2 = e['i1'], e['i2']
            u1 = triuprightn(verts, tris[tis[0]], i1, i2)
            u2 = triuprightn(verts, tris[tis[1]], i1, i2)
            cosangle = np.dot(u1, u2)
            if cosangle <= discard_cosangle:
                e['tis'] = None
            else:
                e['angle'] = math.tau - math.acos(cosangle)
        else:
            print('Edge between v' + str(e['i1']) + ' and v' + str(e['i2']) +
                ' has ' + str(len(tis)) + ' triangles attached, not supported!')
            e['tis'] = None
    edges = [e for e in edges if e['tis'] is not None]
    return edges

def cube():
    verts = [[-1, -1, -1], [-1, -1, 1], [-1, 1, -1], [-1, 1, 1], 
        [1, -1, -1], [1, -1, 1], [1, 1, -1], [1, 1, 1]]
    tris = [[0, 1, 2], [1, 2, 3], [0, 1, 5], [0, 5, 4], 
        [0, 2, 6], [0, 6, 4], [4, 5, 6], [5, 6, 7], 
        [2, 3, 7], [2, 7, 6], [1, 3, 7], [1, 7, 5]]
    edges = mesh_edges(verts, tris)
    # import pprint # debugging only
    # print('Edges for cube: ')
    # pprint.pprint(edges)
    return verts, tris, edges

def nomesh():
    return [], [], []

def plane():
    verts = [[-1, -1, 0], [-1, 1, 0], [1, -1, 0], [1, 1, 0]]
    tris = [[0, 1, 2], [1, 2, 3]]
    edges = mesh_edges(verts, tris)
    return verts, tris, edges
    
def plane_one_edge():
    verts = [[-1, -1, 0], [-1, 1, 0], [1, -1, 0], [1, 1, 0]]
    tris = [[0, 1, 2], [1, 2, 3]]
    edges = [{'i1': 0, 'i2': 1, 'tis': [0], 'angle': math.tau}]
    return verts, tris, edges
    
def octagon():
    l = tan(pi / 8.0)
    verts = [[0, 0, 0], 
        [1, l, 0], [l, 1, 0], [-l, 1, 0], [-1, l, 0],
        [-1, -l, 0], [-l, -1, 0], [l, -1, 0], [1, -l, 0]]
    tris = [[0, 1, 2], [0, 2, 3], [0, 3, 4], [0, 4, 5],
        [0, 5, 6], [0, 6, 7], [0, 7, 8], [0, 8, 1]]
    edges = mesh_edges(verts, tris)
    return verts, tris, edges

def halfplane(sz):
    verts = [[0, -sz, 0], [0, sz, 0], [0, -sz, -sz], [0, sz, -sz]]
    tris = [[0, 1, 2], [1, 2, 3]]
    edges = [{'i1': 0, 'i2': 1, 'tis': [0], 'angle': math.tau}]
    return verts, tris, edges
    
def halfplane_btm(x, z):
    big = 50
    verts = [[x, -big, z], [x, big, z], [big, -big, z], [big, big, z]]
    tris = [[0, 1, 2], [1, 2, 3]]
    edges = [{'i1': 0, 'i2': 1, 'tis': [0], 'angle': math.tau}]
    return verts, tris, edges
    
def longgap(sz):
    big = 50
    verts = [[-sz, -big, 0], [-sz, big, 0], [-big, -big, 0], [-big, big, 0],
        [sz, -big, 0], [sz, big, 0], [big, -big, 0], [big, big, 0]]
    tris = [[0, 1, 2], [1, 2, 3], [4, 5, 6], [5, 6, 7]]
    edges = [{'i1': 0, 'i2': 1, 'tis': [0], 'angle': math.tau},
        {'i1': 4, 'i2': 5, 'tis': [2], 'angle': math.tau}]
    return verts, tris, edges

def infwedge(angle, sz):
    assert(angle < math.pi)
    x = sz * math.cos(angle/2.0)
    z = sz * math.sin(angle/2.0)
    verts = [[0, -sz, 0], [0, sz, 0], [-x, -sz, z], [-x, sz, z], [-x, -sz, -z], [-x, sz, -z]]
    tris = [[0, 1, 2], [1, 2, 3], [0, 1, 4], [1, 4, 5]]
    edges = [{'i1': 0, 'i2': 1, 'tis': [0, 2], 'angle': math.tau - angle}]
    return verts, tris, edges

def halfplane_nonorthog(angle, sz):
    assert(angle < math.pi)
    x = sz * math.cos(angle/2.0)
    z = sz * math.sin(angle/2.0)
    verts = [[0, -sz, 0], [0, sz, 0], [-x, -sz, z], [-x, sz, z]]
    tris = [[0, 1, 2], [1, 2, 3]]
    edges = [{'i1': 0, 'i2': 1, 'tis': [0], 'angle': math.tau}]
    return verts, tris, edges
