#Offline diffraction simulation scripts from pisha2020approximate
#Copyright (C) 2020 Louis Pisha <lpisha@eng.ucsd.edu> / Sonic Arts Research & Development <sonicarts.ucsd.edu> / UC San Diego / The Regents of the University of California 
#See LICENSE.txt for licensing information

# Volumetric Diffraction and Transmission - Introduced in pisha2020approximate
# Portions of the VDaT algorithm may be patented.

from math import *
import numpy as np
from numpy import exp2

from common import *

class VDaT_BTM_Apx():
    def __init__(self, 
            o1=1.44299654, o2=0.91902257, o3=-0.32802643, o4=1.30756273, o5=-0.33781434, h1=1.58287113,
            k1=0.77353521, k2=1.04458582, k3=0.79911644, k4=-1.0509826, k5=0.68170791, k6=0.34534812,
            coherence=True):
        self.o1, self.o2, self.o3, self.o4, self.o5 = o1, o2, o3, o4, o5
        self.h1 = h1
        self.k1, self.k2, self.k3, self.k4, self.k5, self.k6 = k1, k2, k3, k4, k5, k6
        self.coherence = coherence

    def openlfmag(self, ztotal, rout):
        # There were originally multiplicative parameters on ztotal and rout
        # here, which attained the values 0.3635545 and 0.36856452, within
        # about 1% of each other, so they were removed and their joint effect
        # subsumed into o2 and o3 (with further optimization)
        a = 1.0 / sqrt(ztotal*ztotal + rout*rout)
        rat = self.o1 * ztotal / rout
        b = self.o2 + self.o3 / (rat + 1.0 / rat)
        c = 1.0 + pow(self.o4 * ztotal / rout, self.o5)
        return a * b * c

    def holemag(self, ztotal, rout):
        # The 2.0 here was a parameter that attained the value 2.00081305
        return min(self.h1 * rout / pow(ztotal, 2.0), 1.0 / ztotal)

    def knee(self, ztotal, rout):
        z_avg = ztotal * 0.5
        base = log10(345.0 * pow(z_avg, self.k6) / rout)
        k = self.k1 * pow(base - self.k5, self.k2) + self.k3 * base + self.k4
        return pow(10.0, k)

    def freqs_ratio(self, freqs, ztotal, rout):
        return (freqs + 1e-3) / self.knee(ztotal, rout)

    def dring_inside(self, freqs, ztotal, rout):
        m_open = self.openlfmag(ztotal, rout)
        fratio = self.freqs_ratio(freqs, ztotal, rout)
        # magmult = sqrt(freqmult)
        return np.minimum(1.0, 0.5 / np.sqrt(fratio)) * m_open

    def dring_outside(self, freqs, ztotal, rout):
        m_direct = 1.0 / ztotal
        m_hole = self.holemag(ztotal, rout)
        fratio = self.freqs_ratio(freqs, ztotal, rout)
        return np.maximum(np.minimum(fratio * m_direct, m_direct), m_hole)

    def lfcoherence(self, freqs, nfd, fl, fh):
        # The factor is computed so that if we had coherence in all bands, it would
        # double the amplitude at low freqency.
        #cohfact = pow(2.0, 1.0 / nfd)
        #log2(cohfact) = 1.0 / nfd
        # Linear-in-loglog fade down from cohfact at fl to 1.0 at fh
        freqs = np.where(freqs == 0.0, 0.001, freqs)
        fadelg = np.log2(freqs / fh) / (nfd * log2(fl / fh))
        fadelg = np.minimum(1.0 / nfd, np.maximum(0.0, fadelg))
        return np.exp2(fadelg)

    def rxm_to_resp(self, freqs, ringdefs, rxm, ztotal):
        openresp = np.ones_like(freqs) / ztotal
        ret = np.zeros_like(freqs) if rxm[-1] == 0 else np.array(openresp)
        for fi in range(len(ringdefs)-2, -1, -1):
            rout = ringdefs[fi]['r']
            xmin = rxm[fi+1]
            xmout = rxm[fi]
            dxm = abs(xmout - xmin)
            if xmin > xmout:
                dring = self.dring_outside(freqs, ztotal, rout)
                ret -= dxm * (openresp - dring)
            elif xmin < xmout:
                dring = self.dring_inside(freqs, ztotal, rout)
                ret += dxm * dring
        for fi in range(len(ringdefs)-1):
            if self.coherence and rxm[fi+1] == rxm[fi] and rxm[fi] != 0.0 and rxm[fi] != 1.0:
                ret *= self.lfcoherence(freqs, len(ringdefs)-1, ringdefs[fi]['f'], ringdefs[fi+1]['f'])
        return ret


def vdat_2d_rxm(x, ringdefs):
    rxm = []
    for r in ringdefs:
        count = 0
        nangles = r['nangles']
        for n in range(nangles):
            psi = tau * n / nangles + 0.1
            xv = cos(psi) * r['r']
            if xv < x:
                count += 1
        rxm.append(float(count) / nangles)
    return rxm

def vdat(verts, tris, edges, ps, pk, ringdefs, freqs, always, vdatparams, btmapx=None):
    FLOAT_THRESHOLD = 1e-7
    ps = np.array(ps)
    pk = np.array(pk)
    seglen = norm(ps - pk)
    if seglen < FLOAT_THRESHOLD:
        raise ValueError('Source and sink on top of each other')
    rxm = [0.0] * len(ringdefs)
    ple = None
    draw_subpaths = []
    # VDaT plane
    def compute_vdat_plane():
        n = (pk - ps) / seglen
        if abs(n[0]) >= 0.1:
            u = np.array([-n[2] / n[0], 0.0, 1.0])
        elif abs(n[1]) >= 0.1:
            u = np.array([0.0, -n[2] / n[1], 1.0])
        else:
            u = np.array([1.0, 0.0, -n[0] / n[2]])
        u = normalize(u)
        v = np.cross(n, u)
        return u, v
    plane_u, plane_v = compute_vdat_plane()
    # Geometry functions
    def subsegment_xm(s, e):
        openval = 1.0 if always else 2.0
        return openval if not SegIntersectAnyTriangle(verts, tris, s, e) else 0.0
    def cylinder_xm(rad, nangles, showsps):
        xm = 0.0
        oarc_count, oarc_max, oarc_first_count, oarc_is_first = 0, 0, 0, True
        sps = []
        for n in range(nangles):
            psi = 0.1 + tau * n / nangles
            qs = ps + plane_u * rad * cos(psi) + plane_v * rad * sin(psi)
            qk = pk + plane_u * rad * cos(psi) + plane_v * rad * sin(psi)
            def subpath_xm(sp):
                D = 1.0
                for i in range(len(sp)-1):
                    D *= subsegment_xm(sp[i], sp[i+1])
                if showsps:
                    sps.append({'sp': sp, 'xm': D})
                return D
            Da = 0.0
            Nsp = 0
            if vdatparams['subpath0']:
                Da += subpath_xm([ps, qs, qk, pk])
                Nsp += 1
            if vdatparams['subpath1']:
                Da += subpath_xm([ps, qs, pk])
                Da += subpath_xm([ps, qk, pk])
                Nsp += 2
            if vdatparams['subpath2']:
                Da += subpath_xm([ps, (qs + qk) * 0.5, pk])
                Nsp += 1
            if vdatparams['subpath3']:
                Da += subpath_xm([ps, (qs + 0.5 * (qs - qk)), (qk + 0.5 * (qk - qs)), pk])
                Nsp += 1
            xm_a = Da / Nsp
            if xm_a >= 1.0:
                oarc_count += 1
                if oarc_is_first: oarc_first_count += 1
            else:
                oarc_count = 0
                oarc_is_first = False
            oarc_max = max(oarc_max, oarc_count)
            xm += xm_a
        xm = min(xm / nangles, 1.0)
        oarc_max = max(oarc_max, oarc_first_count + oarc_count)
        return xm, tau * oarc_max / nangles, sps
    # Direct path
    fi = len(ringdefs)-1
    assert(ringdefs[fi]['nangles'] <= 1)
    xm = subsegment_xm(ps, pk)
    rxm[fi] = min(1.0, xm)
    if ringdefs[fi]['showsps']:
        draw_subpaths.append({'sp': [ps, pk], 'xm': xm})
    if xm > 0.0: #Direct path not blocked
        if not always:
            print('Direct path not blocked')
            rxm = [1.0 / seglen] * len(ringdefs)
            return rxm, np.ones_like(freqs) / seglen, [seglen], draw_subpaths
        ple = seglen
    fi -= 1
    # Normal paths
    while fi >= 0:
        rad = ringdefs[fi]['r']
        assert(rad > 0)
        rxm[fi], oarc, sps = cylinder_xm(rad, ringdefs[fi]['nangles'], ringdefs[fi]['showsps'])
        draw_subpaths += sps
        if ple is None and oarc > 0:
            #print('band {:.1f} oarc={:.4f}'.format(ringdefs[fi]['f'], oarc))
            r_est = max(rad * cos(oarc * 0.5), ringdefs[fi+1]['r'])
            ple = sqrt(4.0 * r_est * r_est + seglen * seglen)
        fi -= 1
    if ple is None:
        ple = seglen
    # Approximate BTM
    if btmapx is None:
        btmapx = VDaT_BTM_Apx()
    vmag = btmapx.rxm_to_resp(freqs, ringdefs, rxm, seglen)
    return rxm, vmag, [ple], draw_subpaths
    
