#Offline diffraction simulation scripts from pisha2020approximate
#Copyright (C) 2020 Louis Pisha <lpisha@eng.ucsd.edu> / Sonic Arts Research & Development <sonicarts.ucsd.edu> / UC San Diego / The Regents of the University of California 
#See LICENSE.txt for licensing information

#[X] Theory of Diffraction
#where [X] = 'Geometrical' or 'Uniform', see below

from common import *
from math import *
import cmath

def kouyoumjianlenatten(dinc, ddiff, k):
    num = cmath.exp(-1j * k * (dinc + ddiff)) 
    denom = sqrt(dinc * ddiff * (dinc + ddiff))
    return num / denom


#Geometrical Theory of Diffraction
#Based on keller1962geometrical

def gtd_core(thetai, li, thetad, ld, v, phii, k, softboundary):
    A = kouyoumjianlenatten(li, ld, k)
    frontfactor = A * cmath.exp(-1j * pi / 4.0) * v * sin(v * pi) / (sqrt(2.0 * pi * k) * sin(phii))
    def B(beta):
        Binv = cos(v * pi) - cos(v * beta)
        if Binv == 0.0:
            return inf
        return 1.0 / Binv
    return frontfactor * (B(thetad - thetai) - pm(softboundary, B(thetad + thetai)))


#Uniform Theory of Diffraction
#Based on appendix A of tsingos2001modeling
#(NOTE THAT APPENDIX A OF tsingos2001modeling HAS MULTIPLE MISPRINTS)
#which is based on kouyoumjian1974uniform, mcnamara1990introduction, and
#kawai1981sound.

#thetai, thetad: \phi' and \phi in 1974, \alpha_i and \alpha_d in 2001
#li, ld: s' and s in 1974, \rho and r in 2001
#phii: \beta_0 in 1974, \theta_i in 2001
def utd_core(thetai, li, thetad, ld, v, phii, k, softboundary):
    #print('utd_core: thetai {:.0f}, li {:.3f}, thetad {:.0f}, ld {:.3f}, v {:.3f}, phii {:.0f}, k {:.3f}'.format(
    #    degrees(thetai), li, degrees(thetad), ld, v, degrees(phii), k))
    A = kouyoumjianlenatten(li, ld, k)
    frontfactor = -A * v * cmath.exp(-1j * pi / 4.0) / (2.0 * sqrt(2.0 * pi * k) * sin(phii))
    Lambda = li * ld * (sin(phii) ** 2) / (li + ld)
    n = 1.0 / v
    def sign(x):
        return 1.0 if x > 0.0 else -1.0 if x <= 0.0 else 0.0
    def equationhalf(beta):
        def equationquarter(beta, plus):
            def Npm(plus, beta):
                if plus:
                    return 0.0 if beta <= pi * (n - 1.0) else 1.0
                else:
                    return (-1.0 if beta < pi * (1.0 - n) else
                        1.0 if beta > pi * (1.0 + n) else 0.0)
            betaarg = beta - (2.0 * pi * n * Npm(plus, beta))
            cotarg = 0.5 * v * (pi + pm(plus, beta))
            if abs(cotarg) < 0.01: #Going to be singular
                #print('Almost singular mode')
                eps = pi + pm(plus, betaarg)
                #assert(abs(betaarg - pm(plus, pi - eps)) < 1e-6)
                assert(abs(eps) < 1.0)
                return (n * cmath.exp(1j * pi / 4.0) *
                    (sqrt(2.0 * pi * k * Lambda) * sign(eps) - 
                    2.0 * k * Lambda * eps * cmath.exp(1j * pi / 4.0)))
            else:
                #print('Not almost singlar')
                a = 2.0 * (cos(betaarg / 2.0) ** 2)
                X = k * Lambda * a
                if X < 0.8:
                    F = sqrt(pi * X) * (1.0 - sqrt(X)/(0.7*sqrt(X) + 1.2)) 
                else:
                    F = 1.0 - 0.8/((X + 1.25) ** 2)
                F *= cmath.exp(1j * (pi / 4.0) * (1.0 - sqrt(X)/(X+1.4)))
                return F / tan(cotarg)
        return equationquarter(beta, True) + equationquarter(beta, False)
    D = equationhalf(thetad - thetai) - pm(softboundary, equationhalf(thetad + thetai))
    return frontfactor * D

# For both

def call_xtd(xtd_core, verts, tris, edges, ps, pk, freqs, c, always, xtd_oob):
    xm_ret = [0.0] * len(freqs)
    pathlens = []
    for edge in edges:
        ri, rd, zi, zd, li, ld, phii, apex, inedge, zedgelo, zedgehi, thetai, thetad, v \
            = edge_params(verts, tris, edge, ps, pk)
        if thetad - thetai < pi and not always:
            # Not in shadow region of this edge
            continue
        if not inedge and not xtd_oob:
            # Diffraction apex point out of bounds of edge
            continue
        for fi, f in enumerate(freqs):
            xm = xtd_core(thetai, li, thetad, ld, v, phii, tau * f / c, False)
            if not always: xm *= 2.0 #TODO better renormalization
            xm_ret[fi] += xm
        pathlens.append(li + ld)
    # Direct path
    if not SegIntersectAnyTriangle(verts, tris, ps, pk):
        print('XTD direct path open')
        directlen = norm(np.array(pk) - np.array(ps))
        pathlens.append(directlen)
        for fi, f in enumerate(freqs):
            xm_ret[fi] += cmath.exp(-1j * tau * f * directlen / c) / directlen
    # Done
    xm_ret = [abs(x) for x in xm_ret]
    return (xm_ret, pathlens)
